"""Extra operators."""

from __future__ import annotations

import asyncio
import builtins

from typing import TypeVar, Awaitable, Callable, AsyncIterable, AsyncIterator, Any

from .combine import amap, smap
from ..core import pipable_operator

__all__ = ["action", "print"]


T = TypeVar("T")


@pipable_operator
def action(
    source: AsyncIterable[T],
    func: Callable[[T], Awaitable[Any] | Any],
    ordered: bool = True,
    task_limit: int | None = None,
) -> AsyncIterator[T]:
    """Perform an action for each element of an asynchronous sequence
    without modifying it.

    The given function can be synchronous or asynchronous.

    The results can either be returned in or out of order, depending on
    the corresponding ``ordered`` argument. This argument is ignored if the
    provided function is synchronous.

    The coroutines run concurrently but their amount can be limited using
    the ``task_limit`` argument. A value of ``1`` will cause the coroutines
    to run sequentially. This argument is ignored if the provided function
    is synchronous.
    """
    if asyncio.iscoroutinefunction(func):

        async def ainnerfunc(arg: T, *_: object) -> T:
            awaitable = func(arg)
            assert isinstance(awaitable, Awaitable)
            await awaitable
            return arg

        return amap.raw(source, ainnerfunc, ordered=ordered, task_limit=task_limit)

    else:

        def innerfunc(arg: T, *_: object) -> T:
            func(arg)
            return arg

        return smap.raw(source, innerfunc)


@pipable_operator
def print(
    source: AsyncIterable[T],
    template: str = "{}",
    sep: str = " ",
    end: str = "\n",
    file: Any | None = None,
    flush: bool = False,
) -> AsyncIterator[T]:
    """Print each element of an asynchronous sequence without modifying it.

    An optional template can be provided to be formatted with the elements.
    All the keyword arguments are forwarded to the builtin function print.
    """

    def func(value: T) -> None:
        string = template.format(value)
        builtins.print(
            string,
            sep=sep,
            end=end,
            file=file,
            flush=flush,
        )

    return action.raw(source, func)



"""Gather the pipe operators."""

from __future__ import annotations

from . import stream

accumulate = stream.accumulate.pipe
action = stream.action.pipe
amap = stream.amap.pipe
chain = stream.chain.pipe
chunks = stream.chunks.pipe
concat = stream.concat.pipe
concatmap = stream.concatmap.pipe
cycle = stream.cycle.pipe
delay = stream.delay.pipe
dropwhile = stream.dropwhile.pipe
enumerate = stream.enumerate.pipe
filter = stream.filter.pipe
flatmap = stream.flatmap.pipe
flatten = stream.flatten.pipe
getitem = stream.getitem.pipe
list = stream.list.pipe
map = stream.map.pipe
merge = stream.merge.pipe
print = stream.print.pipe
reduce = stream.reduce.pipe
skip = stream.skip.pipe
skiplast = stream.skiplast.pipe
smap = stream.smap.pipe
spaceout = stream.spaceout.pipe
starmap = stream.starmap.pipe
switch = stream.switch.pipe
switchmap = stream.switchmap.pipe
take = stream.take.pipe
takelast = stream.takelast.pipe
takewhile = stream.takewhile.pipe
timeout = stream.timeout.pipe
until = stream.until.pipe
zip = stream.zip.pipe
ziplatest = stream.ziplatest.pipe



"""Utilities for testing stream operators."""

from __future__ import annotations

import asyncio
from collections import deque
from contextlib import contextmanager
from unittest.mock import Mock
from typing import (
    TYPE_CHECKING,
    Awaitable,
    Callable,
    List,
    Protocol,
    TypeVar,
    AsyncIterable,
    AsyncIterator,
    ContextManager,
    Iterator,
    cast,
    Any,
)

import pytest

from .core import StreamEmpty, streamcontext, pipable_operator

if TYPE_CHECKING:
    from _pytest.fixtures import SubRequest
    from aiostream.core import Stream

__all__ = ["add_resource", "assert_run", "event_loop_policy", "assert_cleanup"]


T = TypeVar("T")


@pipable_operator
async def add_resource(
    source: AsyncIterable[T], cleanup_time: float
) -> AsyncIterator[T]:
    """Simulate an open resource in a stream operator."""
    loop = asyncio.get_running_loop()
    assert isinstance(loop, TimeTrackingTestLoop)
    try:
        loop.open_resources += 1
        loop.resources += 1
        async with streamcontext(source) as streamer:
            async for item in streamer:
                yield item
    finally:
        try:
            await asyncio.sleep(cleanup_time)
        finally:
            loop.open_resources -= 1


def compare_exceptions(
    exc1: Exception,
    exc2: Exception,
) -> bool:
    """Compare two exceptions together."""
    return exc1 == exc2 or exc1.__class__ == exc2.__class__ and exc1.args == exc2.args


async def assert_aiter(
    source: Stream[object],
    values: List[object],
    exception: Exception | None = None,
) -> None:
    """Check the results of a stream using a streamcontext."""
    results: list[object] = []
    exception_type = (type(exception),) if exception else ()
    try:
        async with streamcontext(source) as streamer:
            async for item in streamer:
                results.append(item)
    except exception_type as exc:
        assert exception is not None
        assert compare_exceptions(exc, exception)
    else:
        assert exception is None
    assert results == values


async def assert_await(
    source: Stream[object],
    values: List[object],
    exception: Exception | None = None,
) -> None:
    """Check the results of a stream using by awaiting it."""
    exception_type = (type(exception),) if exception else ()
    try:
        result = await source
    except StreamEmpty:
        assert values == []
        assert exception is None
    except exception_type as exc:
        assert exception is not None
        assert compare_exceptions(exc, exception)
    else:
        assert result == values[-1]
        assert exception is None


class AssertRunProtocol(Protocol):
    def __call__(
        self, source: Stream[object], values: List[object], exception: Exception | None
    ) -> Awaitable[None]: ...


@pytest.fixture(params=[assert_aiter, assert_await], ids=["aiter", "await"])  # type: ignore[misc]
def assert_run(request: SubRequest) -> AssertRunProtocol:
    """Parametrized fixture returning a stream runner."""
    return cast(AssertRunProtocol, request.param)


@pytest.fixture  # type: ignore[misc]
def event_loop_policy() -> TimeTrackingTestLoopPolicy:
    """Fixture providing a test event loop.

    The event loop simulate and records the sleep operation,
    available as event_loop.steps

    It also tracks simulated resources and make sure they are
    all released before the loop is closed.
    """
    return TimeTrackingTestLoopPolicy()


@pytest.fixture  # type: ignore[misc]
def assert_cleanup() -> Callable[[], ContextManager[TimeTrackingTestLoop]]:
    """Fixture to assert cleanup of resources."""

    def _assert_cleanup() -> ContextManager[TimeTrackingTestLoop]:
        loop = asyncio.get_running_loop()
        assert isinstance(loop, TimeTrackingTestLoop)
        return loop.assert_cleanup()

    return _assert_cleanup


class BaseEventLoopWithInternals(asyncio.BaseEventLoop):
    _ready: deque[asyncio.Handle]
    _run_once: Callable[[], None]


class TimeTrackingTestLoop(BaseEventLoopWithInternals):
    stuck_threshold: int = 100

    def __init__(self) -> None:
        super().__init__()
        self._time: float = 0.0
        self._timers: list[float] = []
        self._selector = Mock()

        self.steps: list[float] = []
        self.open_resources: int = 0
        self.resources: int = 0
        self.busy_count: int = 0

    # Loop internals

    def _run_once(self) -> None:  # type: ignore
        super()._run_once()
        # Update internals
        self.busy_count += 1
        self._timers = sorted(when for when in self._timers if when > self.time())
        # Time advance
        if self.time_to_go:
            when = self._timers.pop(0)
            step = when - self.time()
            self.steps.append(step)
            self.advance_time(step)
            self.busy_count = 0

    def _process_events(self, event_list: object) -> None:
        return

    def _write_to_self(self) -> None:
        return

    # Time management

    def time(self) -> float:
        return self._time

    def advance_time(self, advance: float) -> None:
        if advance:
            self._time += advance

    def call_at(self, when: float, callback: Callable[..., None], *args: Any, **kwargs: Any) -> asyncio.TimerHandle:  # type: ignore
        self._timers.append(when)
        return super().call_at(when, callback, *args, **kwargs)

    @property
    def stuck(self) -> bool:
        return self.busy_count > self.stuck_threshold

    @property
    def time_to_go(self) -> bool:
        return bool(self._timers) and (self.stuck or not self._ready)

    # Resource management

    def clear(self) -> None:
        self.steps = []
        self.open_resources = 0
        self.resources = 0
        self.busy_count = 0

    @contextmanager
    def assert_cleanup(self) -> Iterator[TimeTrackingTestLoop]:
        self.clear()
        yield self
        assert self.open_resources == 0
        self.clear()


class TimeTrackingTestLoopPolicy(asyncio.DefaultEventLoopPolicy):
    _loop_factory = TimeTrackingTestLoop



"""Time-specific operators."""

from __future__ import annotations
import asyncio

from ..aiter_utils import anext
from ..core import streamcontext, pipable_operator

from typing import TypeVar, AsyncIterable, AsyncIterator

__all__ = ["spaceout", "delay", "timeout"]


T = TypeVar("T")


@pipable_operator
async def spaceout(source: AsyncIterable[T], interval: float) -> AsyncIterator[T]:
    """Make sure the elements of an asynchronous sequence are separated
    in time by the given interval.
    """
    timeout = 0.0
    loop = asyncio.get_event_loop()
    async with streamcontext(source) as streamer:
        async for item in streamer:
            delta = timeout - loop.time()
            delay = delta if delta > 0 else 0.0
            await asyncio.sleep(delay)
            yield item
            timeout = loop.time() + interval


@pipable_operator
async def timeout(source: AsyncIterable[T], timeout: float) -> AsyncIterator[T]:
    """Raise a time-out if an element of the asynchronous sequence
    takes too long to arrive.

    Note: the timeout is not global but specific to each step of
    the iteration.
    """
    async with streamcontext(source) as streamer:
        while True:
            try:
                item = await asyncio.wait_for(anext(streamer), timeout)
            except StopAsyncIteration:
                break
            else:
                yield item


@pipable_operator
async def delay(source: AsyncIterable[T], delay: float) -> AsyncIterator[T]:
    """Delay the iteration of an asynchronous sequence."""
    await asyncio.sleep(delay)
    async with streamcontext(source) as streamer:
        async for item in streamer:
            yield item



"""Utilities for asynchronous iteration."""

from __future__ import annotations

import sys
from types import TracebackType

import enum
import warnings
import functools
from typing import (
    TYPE_CHECKING,
    AsyncContextManager,
    AsyncGenerator,
    AsyncIterable,
    Awaitable,
    Callable,
    Type,
    TypeVar,
    AsyncIterator,
    Any,
    cast,
    overload,
)

if TYPE_CHECKING:
    from typing_extensions import ParamSpec

    P = ParamSpec("P")

from contextlib import AsyncExitStack

__all__ = [
    "aiter",
    "anext",
    "await_",
    "async_",
    "is_async_iterable",
    "assert_async_iterable",
    "is_async_iterator",
    "assert_async_iterator",
    "AsyncIteratorContext",
    "aitercontext",
    "AsyncExitStack",
]


# Magic method shorcuts


_UnsetType = enum.Enum("_UnsetType", "UNSET")
UNSET = _UnsetType.UNSET


def aiter(obj: AsyncIterable[T]) -> AsyncIterator[T]:
    """Access aiter magic method."""
    assert_async_iterable(obj)
    return obj.__aiter__()


@overload
def anext(obj: AsyncIterator[T]) -> Awaitable[T]: ...


@overload
def anext(obj: AsyncIterator[T], default: U) -> Awaitable[T | U]: ...


def anext(obj: AsyncIterator[T], default: U | _UnsetType = UNSET) -> Awaitable[T | U]:
    """Access anext magic method."""
    assert_async_iterator(obj)
    if default is UNSET:
        return obj.__anext__()

    async def anext_default_handling_wrapper() -> T | U:
        try:
            return await obj.__anext__()
        except StopAsyncIteration:
            return default

    return anext_default_handling_wrapper()


# Async / await helper functions


async def await_(obj: Awaitable[T]) -> T:
    """Identity coroutine function."""
    return await obj


def async_(fn: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
    """Wrap the given function into a coroutine function."""

    @functools.wraps(fn)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        return await fn(*args, **kwargs)

    return wrapper


# Iterability helpers


def is_async_iterable(obj: object) -> bool:
    """Check if the given object is an asynchronous iterable."""
    return hasattr(obj, "__aiter__")


def assert_async_iterable(obj: object) -> None:
    """Raise a TypeError if the given object is not an
    asynchronous iterable.
    """
    if not is_async_iterable(obj):
        raise TypeError(f"{type(obj).__name__!r} object is not async iterable")


def is_async_iterator(obj: object) -> bool:
    """Check if the given object is an asynchronous iterator."""
    return hasattr(obj, "__anext__")


def assert_async_iterator(obj: object) -> None:
    """Raise a TypeError if the given object is not an
    asynchronous iterator.
    """
    if not is_async_iterator(obj):
        raise TypeError(f"{type(obj).__name__!r} object is not an async iterator")


# Async iterator context

T = TypeVar("T", covariant=True)
Self = TypeVar("Self", bound="AsyncIteratorContext[Any]")
U = TypeVar("U")


class AsyncIteratorContext(
    AsyncIterator[T], AsyncContextManager["AsyncIteratorContext[T]"]
):
    """Asynchronous iterator with context management.

    The context management makes sure the aclose asynchronous method
    of the corresponding iterator has run before it exits. It also issues
    warnings and RuntimeError if it is used incorrectly.

    Correct usage::

        ait = some_asynchronous_iterable()
        async with AsyncIteratorContext(ait) as safe_ait:
            async for item in safe_ait:
                <block>

    It is nonetheless not meant to use directly.
    Prefer aitercontext helper instead.
    """

    _STANDBY = "STANDBY"
    _RUNNING = "RUNNING"
    _FINISHED = "FINISHED"

    def __init__(self, aiterator: AsyncIterator[T]):
        """Initialize with an asynchrnous iterator."""
        assert_async_iterator(aiterator)
        if isinstance(aiterator, AsyncIteratorContext):
            raise TypeError(f"{aiterator!r} is already an AsyncIteratorContext")
        self._state = self._STANDBY
        self._aiterator = aiterator

    def __aiter__(self: Self) -> Self:
        return self

    def __anext__(self) -> Awaitable[T]:
        if self._state == self._FINISHED:
            raise RuntimeError(
                f"{type(self).__name__} is closed and cannot be iterated"
            )
        if self._state == self._STANDBY:
            warnings.warn(
                f"{type(self).__name__} is iterated outside of its context",
                stacklevel=2,
            )
        return anext(self._aiterator)

    async def __aenter__(self: Self) -> Self:
        if self._state == self._RUNNING:
            raise RuntimeError(f"{type(self).__name__} has already been entered")
        if self._state == self._FINISHED:
            raise RuntimeError(
                f"{type(self).__name__} is closed and cannot be iterated"
            )
        self._state = self._RUNNING
        return self

    async def __aexit__(
        self,
        typ: Type[BaseException] | None,
        value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        try:
            if self._state == self._FINISHED:
                return False
            try:
                # No exception to throw
                if typ is None:
                    return False

                # Prevent GeneratorExit from being silenced
                if typ is GeneratorExit:
                    return False

                # No method to throw
                if not hasattr(self._aiterator, "athrow"):
                    return False
                self._aiterator = cast(AsyncGenerator[T, None], self._aiterator)

                # No frame to throw
                if not getattr(self._aiterator, "ag_frame", True):
                    return False

                # Cannot throw at the moment
                if getattr(self._aiterator, "ag_running", False):
                    return False

                # Throw
                try:
                    assert isinstance(self._aiterator, AsyncGenerator)
                    if sys.version_info >= (3, 12):
                        assert value is not None
                        await self._aiterator.athrow(value)
                    else:
                        await self._aiterator.athrow(typ, value, traceback)
                    raise RuntimeError("Async iterator didn't stop after athrow()")

                # Exception has been (most probably) silenced
                except StopAsyncIteration as exc:
                    return exc is not value

                # A (possibly new) exception has been raised
                except BaseException as exc:
                    if exc is value:
                        return False
                    raise
            finally:
                # Look for an aclose method
                aclose = getattr(self._aiterator, "aclose", None)

                # The ag_running attribute has been introduced with python 3.8
                running = getattr(self._aiterator, "ag_running", False)
                closed = not getattr(self._aiterator, "ag_frame", True)

                # A RuntimeError is raised if aiterator is running or closed
                if aclose and not running and not closed:
                    try:
                        await aclose()

                    # Work around bpo-35409
                    except GeneratorExit:
                        pass  # pragma: no cover
        finally:
            self._state = self._FINISHED

    async def aclose(self) -> None:
        await self.__aexit__(None, None, None)

    async def athrow(self, exc: Exception) -> T:
        if self._state == self._FINISHED:
            raise RuntimeError(f"{type(self).__name__} is closed and cannot be used")
        assert isinstance(self._aiterator, AsyncGenerator)
        self._aiterator = cast(AsyncGenerator[T, None], self._aiterator)
        item: T = await self._aiterator.athrow(exc)
        return item


def aitercontext(
    aiterable: AsyncIterable[T],
) -> AsyncIteratorContext[T]:
    """Return an asynchronous context manager from an asynchronous iterable.

    The context management makes sure the aclose asynchronous method
    has run before it exits. It also issues warnings and RuntimeError
    if it is used incorrectly.

    It is safe to use with any asynchronous iterable and prevent
    asynchronous iterator context to be wrapped twice.

    Correct usage::

        ait = some_asynchronous_iterable()
        async with aitercontext(ait) as safe_ait:
            async for item in safe_ait:
                <block>
    """
    aiterator = aiter(aiterable)
    if isinstance(aiterator, AsyncIteratorContext):
        return aiterator
    return AsyncIteratorContext(aiterator)
